/*
** Module   : terminal_win.c
** Abstract : Windows implementation for character mode-based console functions.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130729 Created initial version to implement Windows specific code to handle text
**                  console terminal JNI.
*/

#include "terminal.h"
#include <windows.h>

// Defining our own attribite flags, Windows does not have equivalents
#define ATTR_NORMAL     0x00010000
#define ATTR_STANDOUT   0x00020000
#define ATTR_UNDERLINE  0x00040000
#define ATTR_REVERSE    0x00080000
#define ATTR_BLINK      0x00100000
#define ATTR_DIM        0x00200000
#define ATTR_BOLD       0x00400000
#define ATTR_ALTCHARSET 0x00800000
#define ATTR_INVIS      0x01000000

// pointer to the structure of the interactive terminal
static HANDLE hConsoleIn  = NULL;
static HANDLE hConsoleOut = NULL;
static DWORD consoleInModeInitial  = 0;
static DWORD consoleOutModeInitial = 0;
static CONSOLE_CURSOR_INFO cursorInfoInitial;
// The current state of the console for suspend-resume processing
static DWORD consoleInModeCurrent  = 0;
static DWORD consoleOutModeCurrent = 0;
static CONSOLE_SCREEN_BUFFER_INFO screenBufferInfoCurrent;
static CONSOLE_CURSOR_INFO cursorInfoCurrent;

static int hascolors      = 0; // set when we initialize
static int colors_started = 0; // set when we call start_color().

// Pointer to the currently used Java environment
static JNIEnv * envCurrent = NULL;
// The current terminal name depending on the display caps, but always 4 chars in length plus one
// char for null-terminator
static char termtype[5] = "";

// This is the normal text attribute may be different depending on console initial color setting
static WORD normalTextAttribute =
   FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_INTENSITY | BACKGROUND_BLUE;
// We need to remember the text attribute to be able to restore it before exiting
#define ATTR_UNKNOWN      0xFFFF
static WORD initialTextAttribute = ATTR_UNKNOWN;

// The array of the character attributes to use in text drawing. This is artificial ones.
jint attribute_result[ATTRIBUTE_MAX] =
   {ATTR_NORMAL,     // A_NORMAL
    ATTR_STANDOUT,   // A_STANDOUT
    ATTR_UNDERLINE,  // A_UNDERLINE
    ATTR_REVERSE,    // A_REVERSE
    ATTR_BLINK,      // A_BLINK ??
    ATTR_DIM,        // A_DIM
    ATTR_BOLD,       // A_BOLD
    ATTR_ALTCHARSET, // A_ALTCHARSET
    ATTR_INVIS};     // A_INVIS (is black on black invisible?)

// The array of the special symbols to draw the horizontal and vertical lines and other special
// symbols. This is the Unicode UTF-16LE values which are native for Windows internals.
jint acs_result[ACS_MAX] = {0x250C,  // ASC_ULCORNER
                            0x2514,  // ACS_LLCORNER
                            0x2510,  // ACS_URCORNER
                            0x2518,  // ACS_LRCORNER
                            0x251C,  // ACS_LTEE
                            0x2524,  // ACS_RTEE
                            0x2534,  // ACS_BTEE
                            0x252C,  // ACS_TTEE
                            0x2500,  // ACS_HLINE
                            0x2502,  // ACS_VLINE              
                            0x253C,  // ACS_PLUS
                            0x002D,  // ACS_S1 - dash or small minus
                            0x005F,  // ACS_S9 - underscore
                            0x0004,  // ACS_DIAMOND
                            0x2591,  // ACS_CKBOARD
                            0x00B0,  // ACS_DEGREE
                            0x00B1,  // ACS_PLMINUS
                            0x2022,  // ACS_BULLET
                            0x2191,  // ACS_UARROW
                            0x2193,  // ACS_DARROW
                            0x2190,  // ACS_LARROW
                            0x2192,  // ACS_RARROW,
                            0x00B9}; // ACS_SELECT

// Externals definitions                                                            
// terminal size at the moment of initialization of the library 
extern int base_rows;
extern int base_cols;

// ----------------------------------------
// Common functions used in several modules
/**
 * clearScreen - clear the entire screen. This is replacement for NCURSES functionality.
 *
 * @param   hConsole    handle to console to clean up the screen.
 *
 */ 
void clearScreen(HANDLE hConsole)
{
   COORD coordsIni = {0, 0};        // Initial cursor position 
   DWORD charsWritten;              // The nimber of chars written
   CONSOLE_SCREEN_BUFFER_INFO csbi; // Screen buffer information
   DWORD conSize;                   // The size of console

   // Get the number of character cells in the current buffer. 
   if (!GetConsoleScreenBufferInfo(hConsole, &csbi))
   {
      return;
   }

   conSize = csbi.dwSize.X * csbi.dwSize.Y;

   // Fill the entire screen with blanks.
   if (!FillConsoleOutputCharacter(hConsole,       // Handle to console screen buffer 
                                   (TCHAR) ' ',    // Character to write to the buffer
                                   conSize,        // Number of cells to write 
                                   coordsIni,      // Coordinates of first cell 
                                   &charsWritten)) // Receive number of characters written
   {
      return;
   }

   // Get the current text attribute.
   if (!GetConsoleScreenBufferInfo(hConsole, &csbi))
   {
      return;
   }
   
   // Set the buffer's attributes accordingly.
   if (!FillConsoleOutputAttribute(hConsole,         // Handle to console screen buffer 
                                   csbi.wAttributes, // Character attributes to use
                                   conSize,          // Number of cells to set attribute 
                                   coordsIni,        // Coordinates of first cell 
                                   &charsWritten))   // Receive number of characters written
   {
      return;
   }

   // Put the cursor at its home coordinates.
   SetConsoleCursorPosition(hConsole, coordsIni);
}

// ----------------------------------------
// Local functions used in this module only

/**
 * Extracts the attribute value for the given char to draw.
 *
 * @param   charOut     character to draw, upper 16 bits containing the new attribute value.
 *
 * @return   The new attribute to draw the character in Windows compatible format.
 */
WORD decodeAttribute(int charOut)
{
   // Leave only high half of the int
   charOut &= 0xFFFF0000;
   
   // Initially this is the normal attribute for text rendering
   WORD attrResult = normalTextAttribute;

   // Guess the reverse has priority over underline   
   if ((charOut & ATTR_REVERSE) == ATTR_REVERSE)
   {
      attrResult = BACKGROUND_BLUE | BACKGROUND_GREEN | BACKGROUND_RED;
   }
   // The underline attribute is dependent on the current reverse one because in Windows
   // console there is no actual underlined chars support, reverse to current is used instead. 
   if ((charOut & ATTR_UNDERLINE) == ATTR_UNDERLINE)
   {
      if ((charOut & ATTR_REVERSE) == ATTR_REVERSE)
      {
         attrResult = normalTextAttribute;
      }
      else
      {
         attrResult = BACKGROUND_BLUE | BACKGROUND_GREEN | BACKGROUND_RED;
      }
   }
   attrResult |= (charOut & ATTR_BLINK) ? BACKGROUND_INTENSITY : 0;
   attrResult |= (charOut & ATTR_BOLD) ? FOREGROUND_INTENSITY : 0;
   
   return attrResult; 
}

#define WIN_VK_DOWN        0x28
#define WIN_VK_UP          0x26
#define WIN_VK_LEFT        0x25
#define WIN_VK_RIGHT       0x27
#define WIN_VK_HOME        0x24
#define WIN_VK_F1          0x70
#define WIN_VK_F2          0x71
#define WIN_VK_F3          0x72
#define WIN_VK_F4          0x73
#define WIN_VK_F5          0x74
#define WIN_VK_F6          0x75
#define WIN_VK_F7          0x76
#define WIN_VK_F8          0x77
#define WIN_VK_F9          0x78
#define WIN_VK_F10         0x79
#define WIN_VK_F11         0x7a
#define WIN_VK_F12         0x7b
#define WIN_VK_F13         0x7c
#define WIN_VK_F14         0x7d
#define WIN_VK_F15         0x7e
#define WIN_VK_F16         0x7f
#define WIN_VK_F17         0x80
#define WIN_VK_F18         0x81
#define WIN_VK_F19         0x82
#define WIN_VK_F20         0x83       
#define WIN_VK_DELETE      0x2e
#define WIN_VK_INSERT      0x2d
#define WIN_VK_PAGE_DOWN   0x22
#define WIN_VK_PAGE_UP     0x21
#define WIN_VK_END         0x23

#define NC_VK_DOWN       0402
#define NC_VK_UP         0403 
#define NC_VK_LEFT       0404
#define NC_VK_RIGHT      0405
#define NC_VK_HOME       0406
#define NC_VK_F1         0411
#define NC_VK_F2         0412
#define NC_VK_F3         0413
#define NC_VK_F4         0414
#define NC_VK_F5         0415
#define NC_VK_F6         0416
#define NC_VK_F7         0417
#define NC_VK_F8         0420
#define NC_VK_F9         0421
#define NC_VK_F10        0422
#define NC_VK_F11        0423
#define NC_VK_F12        0424
#define NC_VK_F13        0425
#define NC_VK_F14        0426
#define NC_VK_F15        0427
#define NC_VK_F16        0430
#define NC_VK_F17        0431
#define NC_VK_F18        0432
#define NC_VK_F19        0433
#define NC_VK_F20        0434
#define NC_VK_DELETE     0512
#define NC_VK_INSERT     0513
#define NC_VK_PAGE_DOWN  0522
#define NC_VK_PAGE_UP    0523
#define NC_VK_END        0550

/**
 * Converst the key codes gotten from Windows into NCURSES compatible ones. Works for F* and
 * cursor movement keys.
 *
 * @param   key     key code got from Windows console API calls.
 *
 * @return   The code for respecive keys for NCURSES environment.
 */
int transformToNCURSES(int key)
{
   
   switch (key)
   {
      case WIN_VK_DOWN:
         key = NC_VK_DOWN;
         break;
   
      case WIN_VK_UP:
         key = NC_VK_UP;
         break;
   
      case WIN_VK_LEFT:
         key = NC_VK_LEFT;
         break;
   
      case WIN_VK_RIGHT:
         key = NC_VK_RIGHT;
         break;
   
      case WIN_VK_HOME:
         key = NC_VK_HOME;
         break;
   
      case WIN_VK_F1:
         key = NC_VK_F1;
         break;
   
      case WIN_VK_F2:
         key = NC_VK_F2;
         break;
   
      case WIN_VK_F3:
         key = NC_VK_F3;
         break;
   
      case WIN_VK_F4:
         key = NC_VK_F4;
         break;
   
      case WIN_VK_F5:
         key = NC_VK_F5;
         break;
   
      case WIN_VK_F6:
         key = NC_VK_F6;
         break;
   
      case WIN_VK_F7:
         key = NC_VK_F7;
         break;
   
      case WIN_VK_F8:
         key = NC_VK_F8;
         break;
   
      case WIN_VK_F9:
         key = NC_VK_F9;
         break;
   
      case WIN_VK_F10:
         key = NC_VK_F10;
         break;
   
      case WIN_VK_F11:
         key = NC_VK_F11;
         break;
   
      case WIN_VK_F12:
         key = NC_VK_F12;
         break;
   
      case WIN_VK_F13:
         key = NC_VK_F13;
         break;
   
      case WIN_VK_F14:
         key = NC_VK_F14;
         break;
   
      case WIN_VK_F15:
         key = NC_VK_F15;
         break;
   
      case WIN_VK_F16:
         key = NC_VK_F16;
         break;
   
      case WIN_VK_F17:
         key = NC_VK_F17;
         break;
   
      case WIN_VK_F18:
         key = NC_VK_F18;
         break;
   
      case WIN_VK_F19:
         key = NC_VK_F19;
         break;
   
      case WIN_VK_F20:
         key = NC_VK_F20;
         break;
   
      case WIN_VK_DELETE:
         key = NC_VK_DELETE;
         break;
   
      case WIN_VK_INSERT:
         key = NC_VK_INSERT;
         break;
   
      case WIN_VK_PAGE_DOWN:
         key = NC_VK_PAGE_DOWN;
         break;
   
      case WIN_VK_PAGE_UP:
         key = NC_VK_PAGE_UP;
         break;
   
      case WIN_VK_END:
         key = NC_VK_END;
         break;
   }
   return key;
}

/**
 * Clean up function to restore console defaults when usage is finished.
 */
void terminateConsole()
{
   // Restore the text drawing attribute
   if (hConsoleOut != NULL && initialTextAttribute != ATTR_UNKNOWN)
   {
      // Restore the text attributes was in console before this driver launched
      SetConsoleTextAttribute(hConsoleOut, initialTextAttribute);
      // Restore initial cursor status
      SetConsoleCursorInfo(hConsoleOut, &cursorInfoInitial);
      // Clear client screen before exiting
      clearScreen(hConsoleOut);
   }
   // General system resource deallocation
   FreeConsole();
}

/**
 * Generates error message and throws Java exception.
 *
 * @param   functionName   the name of the function as error source.
 */
void generateException(const char * functionName)
{
   char* spec =
      "System API error in module terminal_win.c, function %s(), GetLastError() code 0x%X.";
   // 1 additional char for null terminator and 8 for DWORD integer in Hex
   char* errmsg = malloc(strlen(spec) + strlen(functionName) + 9); 
   sprintf(errmsg, spec, functionName, GetLastError());
   throwException(envCurrent, "java/lang/IllegalStateException", errmsg);
}

// Local functions used in this module only
// ----------------------------------------

// Public functions start here
/**
 * Windows specific implementation of the moving cursor to the specified position in console.
 * The next character on the console will be placed in this location.
 *
 * @param   yRow     new Y(row) coordinate of the next char to draw.
 *
 * @param   xColumn  new X(column) coordinate of the next char to draw.
 */
void moveNative(int yRow, int xColumn)
{
   COORD newCoord;
   
   newCoord.X = xColumn;
   newCoord.Y = yRow;
   
   if (!SetConsoleCursorPosition(hConsoleOut, newCoord))
   {
      generateException("moveNative");
   }
}

/**
 * Windows specific implementation of the character drawing to the current position in console.
 *
 * @param   chrToDraw     character to draw in the current cell.
 */
void addchNative(int chrToDraw)
{
   static WORD attrCurrent = ATTR_UNKNOWN;
   // Change the text attribute if required
   WORD attr = decodeAttribute(chrToDraw);

   // 0 attribute is also possible meaning black on black
   if (attrCurrent != attr)
   {
      // Change text attribute when it is really neccessary
      if (!SetConsoleTextAttribute(hConsoleOut, attr))
      {
         // Error, exiting
         generateException("addchNative");
         return;
      }
      // Save the currently used attribute
      attrCurrent = attr;
   }
   
   // Put the character to console, the char located at low 16 bit of integer
   // (even when we use the Unicode range)                          
   WORD chr = (WORD)(chrToDraw & 0x0000FFFF);
   DWORD dwChrWtitten = 0;

   // The Unicode version of the output function, explicit usage W suffix   
   if (!WriteConsoleW(hConsoleOut, &chr, 1, &dwChrWtitten, NULL))
   {
      generateException("addchNative");
   }
}      

/**
 * Windows specific implementation of the calculating the color according to the given color pair
 * index.  The new attribute returned will be or-ed with the resulting color.
 *
 * @param   attrInitial     initial value of the attribute to work with.
 *
 * @param   colorPair       the color pair index to be mixed with initial attribute value.
 *
 * @return   The mixed attribute value as result of initial attribute and calculated color.
 */
int colorPairNative(int attrInitial, int colorPair)
{
   int attrResult = attrInitial;

   // This really can be ignored because we never call init_colorpair() so there is no valid
   // colorpair known to the system. At this time we do not support color change outside the
   // schema derived from the current terminal type defaults
   if (colors_started)
   {
//      attrResult |= COLOR_PAIR(colorPair);
   }
   
   return attrResult;
}

/**
 * Beeps the tone on the terminal.
 */
void beepNative()
{
   // Simple alert beep
   MessageBeep(0xFFFFFFFF);
}

/**
 * Clears the current console screen.
 */
void clearNative()
{
   // Clear terminal screen
   clearScreen(hConsoleOut);
}

/**
 * Returns the safe termname without null string on return.
 *
 * @return   The terminal type or "" on any failure.
 */
char* safe_termname()
{
   // CO80, BW80 or MONO depending on display caps
   return termtype;
}

/**
 * Get X coordinate of the cursor.
 *
 * @return   The integer value of the X coordinate.
 */
jint getx()
{
   CONSOLE_SCREEN_BUFFER_INFO csbInfo;
   
   if (GetConsoleScreenBufferInfo(hConsoleOut, &csbInfo))
   {
      return (jint) csbInfo.dwCursorPosition.X;
   }
   else
   {
      generateException("getx");
      return -1;
   }
}

/**
 * Get Y coordinate of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   jo      reference to the Java object who called this function.
 *
 * @return   The integer value of the Y coordinate.
 */
jint gety()
{
   CONSOLE_SCREEN_BUFFER_INFO csbInfo;
   
   if (GetConsoleScreenBufferInfo(hConsoleOut, &csbInfo))
   {
      return (jint) csbInfo.dwCursorPosition.Y;
   }
   else
   {
      generateException("gety");
      return -1;
   }
}

/**
 * Check if the terminal is able to display colors or not.
 *
 * @return   <code>true</code> if display is color capable, <code>false</code> otherwise.
 */
jboolean hasColors()
{
   // This depends on the currently used terminal type, see initConsole()
   return hascolors;
}

/**
 * This method performs console initialization on the native side.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 */
void initConsole(JNIEnv *env, char * termname)
{
   // This is mandatory the first action to do with terminal to make error processing working
   envCurrent = env;
   // Paranoid error checking has been implemented here
   jboolean initCompleted = JNI_TRUE;
   // This is failed if we already has the console window
   AllocConsole();
   // Getting the standard handles
   hConsoleIn  = GetStdHandle(STD_INPUT_HANDLE);
   hConsoleOut = GetStdHandle(STD_OUTPUT_HANDLE);
   initCompleted = hConsoleIn != NULL && hConsoleOut != NULL;
   hascolors = JNI_TRUE;
   // Getting the terminal type
   HDC hdcDesktop = GetDC(NULL); // Obtain DC for whole screen
   if (hdcDesktop != NULL)
   {
      if (GetDeviceCaps(hdcDesktop, BITSPIXEL) > 8)
      {
         // We have 16, 24 or 32 bpp color model for display
         strcpy(termtype, "CO80");
      }
      else if (GetDeviceCaps(hdcDesktop, NUMCOLORS) < 16)
      {
         // Assuming 2 possible colors means monochrome display like Hercules
         strcpy(termtype, "MONO");
         hascolors = JNI_FALSE;
      }
      else
      {
         // Assuming 16 or 256 palette colors means black and white
         strcpy(termtype, "BW80");
         hascolors = JNI_FALSE;
      }
   }
   else
   {
      // This is default when the desktop DC is not available
      strcpy(termtype, "CO80");
   }
   if (!hascolors)
   {
      // Default color attribute for MONO or BW80 terminals - bright on black
      normalTextAttribute = FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED |
                            FOREGROUND_INTENSITY;
   }
   // Save initial console modes
   if (initCompleted)
   {
      initCompleted = GetConsoleMode(hConsoleIn, &consoleInModeInitial);
   }
   if (initCompleted)
   {
      initCompleted = GetConsoleMode(hConsoleOut, &consoleOutModeInitial);
   }
   // Set new console modes
   if (initCompleted)
   {
      initCompleted = SetConsoleMode(hConsoleIn, 0);
   }
   if (initCompleted)
   {
      initCompleted = SetConsoleMode(hConsoleOut, 0);
   }
   // Initialize statics
   CONSOLE_SCREEN_BUFFER_INFO csbInfo;
   if (initCompleted && GetConsoleScreenBufferInfo(hConsoleOut, &csbInfo))
   {
      base_rows = csbInfo.srWindow.Bottom-csbInfo.srWindow.Top+1;
      base_cols = csbInfo.srWindow.Right-csbInfo.srWindow.Left+1;
      initialTextAttribute = csbInfo.wAttributes;
      if (!hascolors)
      {
         normalTextAttribute = initialTextAttribute;
      }
   }
   else
   {
      initCompleted = JNI_FALSE;
   }
   // Store the initial cursor configuration
   if (initCompleted)
   {
      initCompleted = GetConsoleCursorInfo(hConsoleOut, &cursorInfoInitial);
   }
   // Set console title
   if (initCompleted)
   {
      initCompleted = SetConsoleTitle("Character Client");
   }
   // The final step id to register the session temination routine
   if (initCompleted)
   {
      // Set up the clean function to safe console termination
      atexit((void (*)(void))terminateConsole);
   }
   else
   {
      generateException("initConsole");
   }
}

/**
 * This method reads the key from keyboard by the special reading thread from Windows console.
 *
 * @return   The integer code of the key that was read or -1 if there is no character to read.
 */
jint readKey()
{
   INPUT_RECORD inRec;
   DWORD numberOfRecRead = 0;
   
   if (ReadConsoleInput(hConsoleIn, &inRec, 1, &numberOfRecRead))
   {
      if (numberOfRecRead > 0 &&
          inRec.EventType == KEY_EVENT &&
          inRec.Event.KeyEvent.bKeyDown)
      {
         // We need this call to disable event passing to parent console
         if (FlushConsoleInputBuffer(hConsoleIn))
         {
            // If there is a char, get it otherwise get virtual key
            if (inRec.Event.KeyEvent.uChar.AsciiChar != 0)
            {
               return (jint) inRec.Event.KeyEvent.uChar.AsciiChar;
            }
            else
            {
               // Special transformation required here but only for keys without
               // char value defined.
               return (jint) transformToNCURSES(inRec.Event.KeyEvent.wVirtualKeyCode);
            }
         }
         else
         {
            generateException("flush buffer in readKey");
         }
      }
   }
   else
   {
      generateException("readKey");
   }

   return (jint) -1;
}

/**
 * Reset terminal settings.  May be called only if suspend() was previously called. 
 */
void resetMode()
{
   // Return code
   jboolean ret = JNI_FALSE;
   // And restore the modes
   if (SetConsoleMode(hConsoleIn, consoleInModeCurrent))
   {
      ret = SetConsoleMode(hConsoleOut, consoleOutModeCurrent);
   }
   // Error encountered during mode resetting
   if (!ret)
   {
      generateException("resetMode");
   }
}

/**
 * Re-activates the console mode after ending of the previous suspension state.
 */
void resume()
{
   // Return code
   jboolean ret = JNI_FALSE;
   // Return to the previous tty mode stored locally by suspend() call
   resetMode();
   // Refresh to restore the screen content was before suspension
   ret = SetConsoleScreenBufferSize(hConsoleOut, screenBufferInfoCurrent.dwSize);
   if (ret)
   {
      ret = SetConsoleCursorPosition(hConsoleOut, screenBufferInfoCurrent.dwCursorPosition);
   }
   if (ret)
   {
      ret = SetConsoleCursorInfo(hConsoleOut, &cursorInfoCurrent);
   }
   // Error encountered during mode resetting
   if (!ret)
   {
      generateException("resume");
   }
}

/**
 * Set absolute cursor position
 *
 * @param   x       new X coordinate of the cursor.
 *
 * @param   y       new Y coordinate of the cursor.
 *
 * @return   <code>true</code> if cursor positioning succeeded.
 */
jboolean setCursorNative(jint x, jint y)
{
   // Set current cursor position
   moveNative(y, x);

   // Check if it was OK
   CONSOLE_SCREEN_BUFFER_INFO csbInfo;
   
   if (GetConsoleScreenBufferInfo(hConsoleOut, &csbInfo))
   {
      return csbInfo.dwCursorPosition.Y == y && csbInfo.dwCursorPosition.X == x;
   }
   else
   {
      generateException("setCursorNative");
      return JNI_FALSE;
   }
}

/**
 * Set the cursor visibility on or off. This method only affects the visibility of the cursor.
 *
 * @param   env     pointer to a structure that provides access to JNI environment.
 *
 * @param   on      when <code>true</code> the cursor will become visible, when <code>false</code>
 *                  the cursor will become invisible.
 *
 * @return   The previous state of the cursor (<code>true</code> means visible).
 */
jboolean setCursorStatus(JNIEnv *env, jboolean on)
{
   jboolean ret = JNI_FALSE;
   
   CONSOLE_CURSOR_INFO ccInfo;
   
   if (GetConsoleCursorInfo(hConsoleOut, &ccInfo))
   {
      // Remember the pevious state
      ret = ccInfo.bVisible;
      // New state to set
      ccInfo.bVisible = on;
      if (SetConsoleCursorInfo(hConsoleOut, &ccInfo))
      {
         // This is the only good exit point, the others - means error
         return ret;
      }
      else
      {
         ret = JNI_FALSE;
      }
   }

   if (!ret)   
   {
      generateException("setCursorStatus");
   }
   
   return ret;
}

/**
 * Simple reset of terminal type.
 *
 * @param   env       pointer to a structure that provides access to JNI environment.
 *
 * @param   termname  The name of the terminal to be restarted.
 */
void restartTerminal(JNIEnv *env, char* termname)
{
   // 4GL in Windows does not allow to change the current terminal type 
   return;
}

/**
 * Save the current state of the console mode and switch to TTY mode temporarily.
 */
void suspend()
{
   jboolean suspendCompleted = JNI_TRUE;
   // Save the current console modes
   suspendCompleted = GetConsoleMode(hConsoleIn, &consoleInModeCurrent);
   if (suspendCompleted)
   {
      suspendCompleted = GetConsoleMode(hConsoleOut, &consoleOutModeCurrent);
   }
   // Save screen buffer info
   if (suspendCompleted)
   {
      suspendCompleted = GetConsoleScreenBufferInfo(hConsoleOut, &screenBufferInfoCurrent);
   }
   // Save cursor info
   if (suspendCompleted)
   {
      suspendCompleted = GetConsoleCursorInfo(hConsoleOut, &cursorInfoCurrent);
   }
   else
   {
      generateException("suspend");
   }
}

/**
 * Synchronizes the state of the console buffers with native terminal window. Meaning the kind of
 * refresh calls depending on the backend used.
 */
void syncNative()
{
   // Looks like the Windows Console API output functions are unbuffered meaning the screen
   // is always in sync.
   fflush(stdout);
}
